import os
import sys
import logging
import re
import functools
import time
import hashlib
import tempfile
from datetime import datetime, timedelta
from pytz import timezone
from typing import Any, Dict, List, Text

from bmlx.flow import Executor, Artifact
from bmlx_components.proto import schema_pb2, model_pb2
from bmlx.utils import import_utils, artifact_utils, io_utils

from bmlx_components.utils import cannon_publisher


class PusherExecutor(Executor):
    def _resolve_model_meta(self, model_meta_path):
        fs, path = io_utils.resolve_filesystem_and_path(model_meta_path)
        if not fs.exists(path):
            raise RuntimeError(
                "model_meta_path %s does not exist!" % model_meta_path
            )

        model_pb = io_utils.parse_pbtxt_file(
            os.path.join(model_meta_path, "converted_model.pbtxt"),
            model_pb2.ConvertedModel(),
        )

        if not (model_pb and model_pb.embedding_path and model_pb.graph_path):
            raise RuntimeError(
                "invalid model meta info parsed from %s" % model_meta_path
            )
        logging.info("parsed pushed model meta info: %s", model_pb)

        fs, path = io_utils.resolve_filesystem_and_path(model_pb.embedding_path)
        if not fs.exists(path):
            raise RuntimeError(
                "model embedding path %s does not exist!"
                % model_pb.embedding_path
            )

        fs, path = io_utils.resolve_filesystem_and_path(model_pb.graph_path)
        if not fs.exists(path):
            raise RuntimeError(
                "model graph path %s does not exist!" % model_pb.graph_path
            )

        return model_pb

    def skip_publish_model(self, model_meta_pb, skip_stale_model_hour):
        if re.match("^[0-9]{8}\/[0-9]{2}$", model_meta_pb.model_version):
            cst_tz = timezone("Asia/Chongqing")
            model_time = datetime.strptime(
                f"{model_meta_pb.model_version} +0800", "%Y%m%d/%H %z"
            )
            if datetime.now(cst_tz) - model_time > timedelta(
                hours=skip_stale_model_hour
            ):
                return True
        return False

    def publish_embeddings(self, exec_properties, model_version, emb_bin_path):
        def get_all_dims(emb_bin_path):
            fs, path = io_utils.resolve_filesystem_and_path(emb_bin_path)
            assert fs.isdir(path)
            emb_dir_list = [f for f in fs.ls(path) if fs.isdir(f)]
            if not emb_dir_list:
                raise RuntimeError("Invalid emb_bin_path %s" % emb_bin_path)

            emb_dims = [int(d.split("/")[-1]) for d in emb_dir_list]
            return emb_dims

        emb_dims = get_all_dims(emb_bin_path)

        emb_collection = exec_properties["emb_collection"]
        for emb_dim in emb_dims:
            input_path = os.path.join(emb_bin_path, str(emb_dim))
            cannon_pub_opts = cannon_publisher.PublishOptions(
                build_type="embed",
                input_path=input_path,
                data_name=exec_properties["model_name"],
                dim=emb_dim,
                version=model_version,
                collection=emb_collection,
                creator=exec_properties["author"],
                namespace=exec_properties["product_namespace"],
                other_args=f"{emb_collection},{emb_dim},{model_version}",
                test_env=exec_properties["test_env"],
            )
            ret = cannon_publisher.publish_resource(cannon_pub_opts)
            if ret < 0:
                raise RuntimeError(
                    "Failed to publish embedding using cannon_publisher, input path: %s"
                    % input_path
                )
            if not cannon_publisher.check_resource(cannon_pub_opts):
                raise RuntimeError("Publish embedding %s failed!" % input_path)
            else:
                logging.info("Publish embedding %s successfully!", input_path)

    def publish_graph(self, exec_properties, model_version, graph_path):
        # publish graph
        cannon_pub_opts = cannon_publisher.PublishOptions(
            build_type="deliver_graph",
            input_path=graph_path,
            data_name=exec_properties["model_name"],
            version=model_version,
            collection=exec_properties["graph_collection"],
            creator=exec_properties["author"],
            namespace=exec_properties["product_namespace"],
            phase="final",
            other_args="send_conf=false;hdfs_move=true",
            test_env=exec_properties["test_env"],
        )
        ret = cannon_publisher.publish_resource(cannon_pub_opts)
        if ret < 0:
            raise RuntimeError("Failed to publish graph using cannon_publisher")
        else:
            logging.info("Publish graph successfully!")

    def save_meta(
        self, meta_output_path, model_name, model_version, origin_model_path
    ):
        logging.info("push finished, model meta gen to %s" % meta_output_path)
        pushed_model = model_pb2.PushedModel()
        pushed_model.version = model_version
        pushed_model.name = model_name
        pushed_model.origin_model_path = origin_model_path
        io_utils.write_pbtxt_file(
            os.path.join(meta_output_path, "pushed_model.pbtxt"), pushed_model
        )

    def execute(
        self,
        input_dict: Dict[Text, List[Artifact]],
        output_dict: Dict[Text, List[Artifact]],
        exec_properties: Dict[Text, Any],
    ):
        self._log_startup(input_dict, output_dict, exec_properties)

        assert (
            "converted_model" in input_dict
            and len(input_dict["converted_model"]) == 1
        )
        model_meta_pb = self._resolve_model_meta(
            input_dict["converted_model"][0].meta.uri
        )
        skip_stale_model_hour = exec_properties["skip_stale_model_hour"]

        if self.skip_publish_model(model_meta_pb, skip_stale_model_hour):
            cst_tz = timezone("Asia/Chongqing")
            logging.info(
                "model is %d hours earlier, model time: %s, "
                "current time: %s, skip to publish this model !!!!",
                skip_stale_model_hour,
                model_meta_pb.model_version,
                datetime.now(cst_tz).strftime("%Y%m%d/%H"),
            )
            return

        model_version = int(time.time())

        # publish embedding
        self.publish_embeddings(
            exec_properties, model_version, model_meta_pb.embedding_path
        )
        self.publish_graph(
            exec_properties, model_version, model_meta_pb.graph_path
        )

        self.save_meta(
            artifact_utils.get_single_uri(output_dict["output"]),
            exec_properties["model_name"],
            model_version,
            os.path.dirname(model_meta_pb.graph_path),
        )
